import uuid 
from .config import BASE_DIR,HTML_FILE_PATH
from collections import OrderedDict

class HtmlShell(object):
    def __init__(self, doc_type: str = 'html', title: str = "Document", charset: str = "UTF-8", http_equiv: str = "X-UA-Compatible", content: str = "IE=edge", name__meta: str = "viewport", language: str = 'en', **kwargs) -> None:
        """
        Represents an HTML shell object.

        Args:
            doc_type (str, optional): The type of the document. Defaults to 'html'.
            title (str, optional): The title of the document. Defaults to "Document".
            charset (str, optional): The character encoding of the document. Defaults to "UTF-8".
            http_equiv (str, optional): The value for the http-equiv attribute. Defaults to "X-UA-Compatible".
            content (str, optional): The value for the content attribute. Defaults to "IE=edge".
            name__meta (str, optional): The value for the name attribute of the meta tag. Defaults to "viewport".
            language (str, optional): The language of the document. Defaults to 'en'.
            **kwargs: Additional keyword arguments.

        """
        self.title = title
        self.content_ = ''
        self.type = doc_type
        if doc_type == 'html':
            self.top = "<!DOCTYPE html>"
            self.html = f"<html lang='{language}'>"

            script: list = kwargs.get('script', [])  # capture direct script
            if script is not None:
                script_box = "".join(f"<script>{str(i)}</script>" for i in script)
            else:
                script_box = ''

            script_from_link: list = kwargs.get('script_from_link', []) or kwargs.get('sfl', [])  # capture scripts from external links
            if script_from_link is not None:
                script_box_from_link = "".join(f"<script src='{str(i)}'></script>" for i in script_from_link)
            else:
                script_box_from_link = ''

            other_links: list = kwargs.get('link_css', [])
            if other_links is not None:
                css = "".join(f"<link rel='stylesheet' href='{str(i)}'>" for i in other_links)  # capture css
            else:
                css = ''

            self.head = f"""
    <head>
        <meta charset="{charset}">
        <meta http-equiv="{http_equiv}" content="{content}">
        <meta name="{name__meta}" content="width=device-width, initial-scale=1.0">
        <title>{title}</title>
        {script_box}{script_box_from_link}{css}
    </head>
                        """

    def __iter__(self):
        """
        Returns an iterator over the HTML shell object.

        Yields:
            str: The HTML shell components.

        """
        buck = [self.top, self.html, self.head, self.content_]
        for i in buck:
            yield i

    def __repr__(self) -> str:
        """
        Returns a string representation of the HTML shell object.

        Returns:
            str: The string representation of the HTML shell object.

        """
        return f"<HtmlShell object: {self.title}>"

    def __str__(self):
        """
        Returns a string representation of the HTML shell object.

        Returns:
            str: The string representation of the HTML shell object.

        """
        return f"<HtmlShell object: {self.title}>"

    def insert(self, obj):
        """
        Inserts content into the HTML shell object.

        Args:
            obj: The content to be inserted.

        Returns:
            HtmlShell: The updated HTML shell object.

        """
        self.content_ = "".join(str(obj))
        return self

    def to_file(self, filepath=HTML_FILE_PATH):
        """
        Writes the HTML shell object to a file.

        Args:
            filepath (str, optional): The path of the file to write to. Defaults to HTML_FILE_PATH.

        Raises:
            ValueError: If the value of 'type' is unknown.

        """
        if self.type == 'html':
            self.all = self.top + "\n" + self.html + "\n" + self.head + '\n' + self.content_ + "\n" + "</html>"
        elif self.type == 'xml':
            self.all = "<html xmlns='http://www.w3.org/1999/xhtml'>"
        else:
            raise ValueError(f"Unknown value '{self.type}' given for argument 'type', this argument receives either 'html' or 'xml' ")
        file_content = self.all
        with open(filepath, 'w') as html:
            html.write(file_content)
    
class HtmlSkeleton(object):
    def __init__(self):
        """
        Initializes an instance of HtmlSkeleton class.
        """
        self.innards = OrderedDict()
    
    def __str__(self) -> str:
        """
        Returns the string representation of the HtmlSkeleton object.
        """
        return self.barebones

    def __repr__(self) -> str:
        """
        Returns the string representation of the HtmlSkeleton object.
        """
        return self.barebones

    def __iter__(self):
        """
        Iterates over the innards of the HtmlSkeleton object.
        """
        for i in self.innards:
            yield i 

    def add(self, obj, name=None):
        """
        Adds an object to the HtmlSkeleton object.

        Args:
            obj: The object to be added.
            name: The name of the object (optional).

        Returns:
            The HtmlSkeleton object.
        """
        if name is None:
            name = str(uuid.uuid4())
        marked_obj = Marker(obj, name)
        self.innards[name] = marked_obj 
        self.finalise()
        return self 
    
    def finalise(self):
        """
        Finalizes the HtmlSkeleton object by joining the innards.
        """
        self.barebones = "".join(i[1].obj+"\n" for i in self.innards.items())
 
    def load_string(self, obj:str):
        """
        Loads a string into the HtmlSkeleton object.

        Args:
            obj: The string to be loaded.
        """
        str_obj = obj 
        self.add(str_obj)
    
    def load_file(self, path:str):
        """
        Loads a file into the HtmlSkeleton object.

        Args:
            path: The path of the file to be loaded.
        """
        with open(path, 'r') as file:
            doc = file.read()
            container = ""
            for i in doc:
                container.join(i)
        self.add(container)
    
    def remove(self, name):
        """
        Removes an object from the HtmlSkeleton object.

        Args:
            name: The name of the object to be removed.
        """
        del self.innards[name]
    
    def wrap(self, wrapper:str, **kwargs):
        """
        Wraps the HtmlSkeleton object with a wrapper tag.

        Args:
            wrapper: The wrapper tag.
            kwargs: Additional attributes for the wrapper tag.

        Returns:
            The wrapped HtmlSkeleton object.
        """
        sb = ''
        for key, item in kwargs.items():
            sb += f"""{key}="{item}" """
        self.barebones = f"<{wrapper} {sb}> \n\t {self.barebones} \n\t </{wrapper}>"
        return self.barebones



    
class Marker(object):
    def __init__(self, object_to_mark, name=None) -> None:
        """
        Initializes a Marker object.

        Args:
            object_to_mark: The object to be marked.
            name: Optional name for the marker.

        Returns:
            None
        """
        self.obj = object_to_mark
        self.name = name 
        if self.name is not None:
            self.name = name
        self.mark()
    
    def mark(self):
        """
        Marks the object with a unique identifier.

        Returns:
            The unique identifier of the marked object.
        """
        if self.name is not None:
            self.id =  (self.obj, self.name)
        else:
            self.id =  (self.obj,)
        return self.id
    
    def __gt__(self, other):
        """
        Compares two Marker objects based on their names.

        Args:
            other: The other Marker object to compare with.

        Returns:
            True if the name of this Marker is greater than the name of the other Marker, False otherwise.

        Raises:
            TypeError: If the names of the two Marker objects are of different types.
        """
        if type(self.name) == type(other.name):
            return self.name > other.name
        else:
            raise TypeError('Cannot compare values of different types')

    def __lt__(self, other):
        """
        Compares two Marker objects based on their names.

        Args:
            other: The other Marker object to compare with.

        Returns:
            True if the name of this Marker is less than the name of the other Marker, False otherwise.

        Raises:
            TypeError: If the names of the two Marker objects are of different types.
        """
        if type(self.name) == type(other.name):
            return self.name < other.name
        else:
            raise TypeError('Cannot compare values of different types')
    
    def __str__(self) -> str:
        """
        Returns a string representation of the Marker object.

        Returns:
            A string representation of the unique identifier of the Marker object.
        """
        return str(self.id) 
    
    def __repr__(self) -> str:
        """
        Returns a string representation of the Marker object.

        Returns:
            A string representation of the unique identifier of the Marker object.
        """
        return str(self.id)

def iwrap(wrapper:str,obj:str, class_:str,**kwargs):
    """
    Wraps the given object with an HTML wrapper element.

    Args:
        wrapper (str): The HTML wrapper element to use.
        obj (str): The object to wrap.
        class_ (str): The CSS class to apply to the wrapper element.
        **kwargs: Additional attributes to add to the wrapper element.

    Returns:
        str: The wrapped object as an HTML string.
    """
    sb = ''
    for key,item in kwargs.items():
        sb += f"""{key}='{item}'"""
    barebones = f"<{wrapper} class='{class_}' {sb}> \n\t {obj} \n\t </{wrapper}>"
    return barebones



set_ = {
    'APP_NAME' :'swiftdeploy',
    'APP_HEADER' : 'swiftdeploy',
    'APP_FOOTER':'swiftdeploy',
}

class Settings:
    pass

config = Settings()
for key, item in set_.items():
    setattr(config, key, item)
css = """:root {
    --primary: #421145;
    --tertiary: rgb(245,245,255);
    --secondary: #9c42a1;
    --text: #5f4e63;
}

body {
    background: var(--tertiary);
    color: var(--text);
}
 


header {
    position: fixed;
    top: 0;
    left: 0;
    width: 100%;
    background: linear-gradient(90deg, var(--primary), var(--secondary));
    color: var(--tertiary);
    padding: 10px;
    box-shadow: 0px 0px 10px 0px rgba(0,0,0,0.75);
    z-index: 9999;
    font-size: 2em;
    text-transform: capitalize;
    align-items: center;
    justify-content: center;
    display: flex;
}

.app-info{
    background: var(--tertiary);
    color: var(--text);
    padding: 20px;
    border-radius: 10px;
    margin-top:45px;
    font-size: 1.5em;
    align-items: center;
    justify-content: center;
    display: flex;
}

form{
    display: flex;
    flex-direction: column;
    align-items: center;
    justify-content: center;
    margin-top: 20px;
    font-size: 1.3em;
    text-transform: capitalize;
}

form input {
    width: 300px;
    padding: 10px;
    margin-bottom: 10px;
    border: 1px solid var(--primary);
    border-radius: 5px;
}

form button {
    width: 150px;
    padding: 10px;
    background-color: var(--primary);
    color: var(--tertiary);
    border: none;
    border-radius: 5px;
    cursor: pointer;
}

form button:hover {
    background-color: var(--secondary);
}
.form-group {
    display: flex;
    justify-content: flex-start;
    align-items: center;
}

.form-group label {
    width: 100px; /* Adjust as needed */
}

.form-group input {
    flex-grow: 1;
}

footer{
    position: fixed;
    bottom: 0;
    left: 0;
    width: 100%;
    background: linear-gradient(90deg, var(--primary), var(--secondary));
    color: var(--tertiary);
    padding: 10px;
    box-shadow: 0px 0px 10px 0px rgba(0,0,0,0.75);
    z-index: 9999;
    font-size: 2em;
    text-transform: capitalize;
    align-items: center;
    justify-content: center;
    display: flex;   
}

"""

config.CSS = css


import os
from swiftdeploy.settings import config
from swiftdeploy.html.htmlcontent import HtmlSkeleton, HtmlShell
from swiftdeploy.html.tags import Division, Header, Tag , Form, FormElement



class MarkupModel:
    """
    This class represents a markup model for a web application.

    Parameters:
    - `model_info` (str): Information to display about the model.
    - `model_func` (function): The function that runs the model and returns the prediction.
    - `form_fields` (dict): A dictionary of input fields the model takes, in the format 'field':'type'.
                            Allowed types include 'text', 'number', 'date', and 'image'.

    Usage:
    1. Import the `ModelApp` class from the `swiftdeploy` module.
    2. Create an instance of `ModelApp`.
    3. Use the instance to interact with the markup model.

    Example:
    ```python
    from swiftdeploy import ModelApp

    modelapp = ModelApp()
    ```

    """

    def __init__(self, **kwargs):
        self.app_name = config.APP_NAME
        self.model_info = kwargs.get("model_info", "")
        self.form_fields: dict = kwargs.get('form_fields')
        self.pagebody = []
        self.model_func = kwargs.get("model_func")
        if self.model_func is None:
            raise ValueError(f"Cannot initialize {self.__class__} without model_func")
        self.makePage()

    def addHeader(self, header_value=config.APP_HEADER):
        """
        Add a header to the markup model.

        Parameters:
        - `header_value` (str): The value of the header.

        """
        header = Header(id=f"app-header", classname=f'app-header', body=header_value).create()
        topdiv = Division(body=self.model_info, classname=f"app-info", id=f"app-info").create()
        self.topdiv_container = HtmlSkeleton()
        self.topdiv_container.add(header)
        self.topdiv_container.add(topdiv)

    def addContent(self, *args):
        """
        Add content to the markup model.

        Parameters:
        - `args` (str): The content to be added.

        Returns:
        - `self`: The instance of the markup model.

        """
        self.content = HtmlSkeleton()
        for item in args:
            self.content.add(item)
        self.content.wrap('div')
        return self

    def addFooter(self, footer_value=config.APP_FOOTER):
        """
        Add a footer to the markup model.

        Parameters:
        - `footer_value` (str): The value of the footer.

        Returns:
        - `self`: The instance of the markup model.

        """
        self.footer = Tag(name='footer', id=f'app-footer', classname=f'app-footer', body=footer_value).create()
        return self

    def createform(self):
        """
        Create a form for the markup model.

        Returns:
        - `form`: The created form.

        Raises:
        - `ValueError`: If form_fields have not been set.

        """
        if self.form_fields is None:
            raise ValueError("form_fields have not been set")
        else:
            form = Form()
            for key, item in self.form_fields.items():
                if item == 'file' or item == 'image':
                    form.add(FormElement(name=f'{key}', id=f"id_{key}", classname=f'class_{key}', input_type=item, label=key, accept='image/*').make())
                else:
                    form.add(FormElement(name=f'{key}', id=f"id_{key}", classname=f'class_{key}', input_type=item, label=key).make())
            return form.set_()

    def compile(self):
        """
        Compile the markup model.

        This method adds the header, content, and footer to the markup model.

        """
        self.addHeader()
        self.addContent(*self.pagebody)
        self.addFooter()

    def makePage(self):
        """
        Create the markup model page.

        This method creates the HTML structure of the markup model page and saves it to a file.

        """
        self.pageObj = HtmlShell(title=config.APP_HEADER, sfl=[], link_css=['./static/index.css'])
        self.body = HtmlSkeleton()
        self.pagebody = [self.createform()]
        self.compile()
        self.body.add(self.topdiv_container.barebones)
        self.body.add(self.content.barebones)
        self.body.add(self.footer)
        self.body.wrap('body')
        self.pageObj.insert(self.body)
        if os.path.exists(os.path.join(config.BASE_DIR, 'templates')):
            self.pageObj.to_file(os.path.join(config.BASE_DIR, 'templates/index.html'))
        else:
            os.makedirs(os.path.join(config.BASE_DIR, 'templates'), exist_ok=True)
            self.pageObj.to_file(os.path.join(config.BASE_DIR, 'templates/index.html'))
        if os.path.exists(os.path.join(config.BASE_DIR, 'static')):
            with open(os.path.join(config.BASE_DIR, 'static/index.css'), 'w') as f:
                f.write(config.CSS)
        else:
            os.makedirs(os.path.join(config.BASE_DIR, 'static'), exist_ok=True)
            with open(os.path.join(config.BASE_DIR, 'static/index.css'), 'w') as f:
                f.write(config.CSS)
        



from setuptools import setup, find_packages

setup(
    name="swiftdeply",
    version="0.5",
    author="Akinrotimi Daniel F",
    author_email="dtenny95@gmail.com",
    description="This package provides a web application wrapper for your machine learning models.",
    long_description=open('README.md', 'r').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/firebreather-heart/swiftdeploy",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'Flask',
    ],
)


import os
from flask import Flask, render_template,request
from flask.views import MethodView
from swiftdeploy.settings import config
from flask import abort

template = os.path.join(config.BASE_DIR, 'templates')
static = os.path.join(config.BASE_DIR, 'static')
webapp = Flask(__name__, template_folder=template, static_folder=static)



         
class WebViews(MethodView):
     """
     Class representing the web views for the application.

     Methods:
     - get: Handles the GET request and renders the 'index.html' template.
     - post: Handles the POST request, validates the input, and returns the model predictions.
     """

     def get(self):
          """
          Handles the GET request and renders the 'index.html' template.

          Returns:
          - The rendered 'index.html' template.
          """
          return render_template('index.html')

     def post(self):
          """
          Handles the POST request, validates the input, and returns the model predictions.

          Returns:
          - The prediction result as an HTML string.
          """
          model_input = [request.form.get(key) for key,_ in config.model.form_fields.items()]
          for i in model_input:
               if i is None:
                    abort(403)  # Return a 403 Forbidden error if any input is missing
               predictions = config.model.model_func(model_input) # Model output is obtained here
          return f"<h1>Prediction: {predictions}</h1>"


webapp.add_url_rule("/",view_func=WebViews.as_view('home'))


from swiftdeploy.html.htmlcontent import HtmlSkeleton,iwrap


class Tag:
    """
    Base class for html tags
    
    Attributes:
        name (str): The name of the HTML tag.
        _class (str): The class attribute of the HTML tag.
        id (str): The id attribute of the HTML tag.
        style (str): The style attribute of the HTML tag.
        body (str): The content of the HTML tag.
        misc (str): Additional tag variables in string format.
        href (str): The href attribute of the HTML tag.
    """
    def __init__(self,body=None,**kwargs) -> None:
        self.name = kwargs.get('name')
        self._class = kwargs.get('classname', '')
        self.id = kwargs.get('id','')
        self.style = kwargs.get('style', '')
        self.body = body #if you want to create an empty tag, then assign empty to body
        self.misc = kwargs.get('misc','') # this is a list containing all other tag variable such as required etc in string format
        self.href = kwargs.get('href', '')
        if self.misc is not None:
            self.misclist = "".join(i+" " for i in self.misc)
        
    def create(self):
        """
        Create an HTML tag based on the provided attributes.

        Returns:
            str: The generated HTML tag.

        Raises:
            None
        """
        if self.body is None:
            return f'<{self.name} href="{self.href}" class="{self._class}" id="{self.id}" {self.misclist}/>'
        elif self.body == 'empty':
            return f'<{self.name} href="{self.href}" class="{self._class}" id="{self.id}" {self.misclist}></{self.name}>'
        else:
            return f'<{self.name} href="{self.href}" class="{self._class}" id="{self.id}" {self.misclist}>{self.body}</{self.name}>'
    
    def __str__(self) -> str:
        return self.create()
    
    def __repr__(self) -> str:
        return self.create()

class FormElement():
    def __init__(self, **kwargs) -> None:
        """
        Initialize a FormElement object.

        Args:
            **kwargs: Additional keyword arguments to customize the FormElement.

        Keyword Args:
            name (str): The name attribute of the form element.
            id (str): The id attribute of the form element.
            classname (str): The class attribute of the form element.
            input_type (str): The type attribute of the input element.
            label (str): The label text for the form element.
            accept (str): The accept attribute for file types.

        Returns:
            None
        """
        self.name = kwargs.get('name','')
        self.id = kwargs.get('id', '')
        self.class_ = kwargs.get('classname', '')
        self.input_type = kwargs.get('input_type', '')
        self.label = kwargs.get('label', '')
        self.accept = kwargs.get('accept', '')  # New attribute for file types
    
    def make(self):
        """
        Generate the HTML tag for the form element.

        Returns:
            str: The HTML tag for the form element.
        """
        if self.input_type == 'file' or self.input_type == 'image':
            self.tag = f""" 
                <label for="{self.name}">{self.label}</label>
                <input name="{self.name}"  class="{self.class_}" type="file" id="{self.id}" accept="{self.accept}"></input>    
            \n"""
        else:
            self.tag = f"""
                <label for="{self.name}">{self.label}</label>
                <input name="{self.name}"  class="{self.class_}" type="{self.input_type}" id="{self.id}"></input>    
            \n"""
        self.tag = iwrap('div', self.tag, class_='form-group')
        return self.tag


class Form(HtmlSkeleton):
    def __init__(self):
        """
        Represents an HTML form element.

        This class provides methods to set attributes and add content to the form element.
        """
        super().__init__()

    def set_(self):
        self.add("""<button type="submit">Submit</button>""")
        return self.wrap("form", enctype="multipart/form-data", method="post")

    def __str__(self) -> str:
        return super().__str__()

    def __repr__(self) -> str:
        return super().__repr__()


class Header(Tag):
    """
    Base class for header tags
    
    Args:
        type_ (int): The header type. If provided, the tag name will be set to 'h{type_}'.
        body (str): The content of the header tag.
        **kwargs: Additional attributes to be added to the header tag.
    """
    def __init__(self, type_=None, body=None, **kwargs) -> None:
        super().__init__(body, **kwargs)
        self.type_ = type_ 
        if isinstance(self.type_, int):
            self.name = f"h{self.type_}"
        else:
            self.name = 'header'
        self.create()

def h_tag(type_=None, body=None, **kwargs):
    """
    Create an HTML header tag with the specified type, body, and additional attributes.

    Args:
        type_ (str): The type of the header tag (e.g., 'h1', 'h2', 'h3', etc.).
        body (str): The content of the header tag.
        **kwargs: Additional attributes to be added to the header tag.

    Returns:
        str: The HTML representation of the header tag.
    """
    param_dict = kwargs
    tag_obj = Header(type_, body, **param_dict).create()
    return tag_obj

class H1(Header):
    """
    sub class for h1 tag
    """
    def __init__(self,body=None, **kwargs) -> None:
        super().__init__(type_=1,  body=body, **kwargs)
        self.type_ = 1 
        self.name = 'h1'
        self.create()

class H2(Header):
    """
    sub class for h2 tag
    """
    def __init__(self,body=None, **kwargs) -> None:
        super().__init__(type_=2,  body=body, **kwargs)
        self.type_ = 2 
        self.name = 'h2'
        self.create()

class H3(Header):
    """
    sub class for h3 tag
    """
    def __init__(self,body=None, **kwargs) -> None:
        super().__init__(type_=3,  body=body, **kwargs)
        self.type_ = 3 
        self.name = 'h3'
        self.create()

class H4(Header):
    """
    sub class for h4 tag
    """
    def __init__(self,body=None, **kwargs) -> None:
        super().__init__(type_=4,  body=body, **kwargs)
        self.type_ = 4 
        self.name = 'h4'
        self.create()

class H5(Header):
    """
    sub class for h5 tag
    """
    def __init__(self,body=None, **kwargs) -> None:
        super().__init__(type_=5,  body=body, **kwargs)
        self.type_ = 5 
        self.name = 'h5'
        self.create()

class H6(Header):
    """
    sub class for h6 tag
    """
    def __init__(self,body=None, **kwargs) -> None:
        super().__init__(type_=6,  body=body, **kwargs)
        self.type_ = 6 
        self.name = 'h6'
        self.create()


        
class Paragraph(Tag):
    """
    Represents an HTML paragraph tag.

    Args:
        body (str): The content of the paragraph.
        **kwargs: Additional attributes to be added to the paragraph tag.

    Attributes:
        name (str): The name of the tag ('p').

    Methods:
        create: Creates the paragraph tag.
    """

    def __init__(self, body=None, **kwargs) -> None:
        super().__init__(body, **kwargs)
        self.name = 'p'
        self.create()

def p_tag(body=None, **kwargs):
    """
    Create a paragraph HTML tag with the given body and optional attributes.

    Args:
        body (str): The content of the paragraph tag.
        **kwargs: Optional attributes to be added to the paragraph tag.

    Returns:
        str: The HTML representation of the paragraph tag.
    """
    param_dict = kwargs 
    tag_obj = Paragraph(body, **param_dict).create()
    return tag_obj

class Division(Tag):
    """
    Represents an HTML <div> tag.

    Args:
        body (str): The content of the <div> tag.
        **kwargs: Additional attributes to be added to the <div> tag.

    Attributes:
        name (str): The name of the tag ('div').

    Methods:
        create: Creates the <div> tag with the specified attributes and content.
    """

    def __init__(self, body=None, **kwargs) -> None:
        super().__init__(body, **kwargs)
        self.name = 'div'
        self.create()
    
def div(body=None, **kwargs):
    param_dict = kwargs
    tag_obj = Division(body, **param_dict).create()
    return tag_obj

def make_tag(name:str,**kwargs):
    """
    Create an HTML tag with the given name and optional attributes.

    Args:
        name (str): The name of the HTML tag.
        **kwargs: Optional attributes for the HTML tag.

    Returns:
        str: The created HTML tag.

    Raises:
        ValueError: If the name parameter is not a string.

    Example:
        >>> make_tag('div', id='myDiv', class_='container')
        '<div id="myDiv" class="container"></div>'
    """
    if not isinstance(name, str):
        raise ValueError(f'expected string object got {type(name)}')
    param_dict = kwargs
    new_tag = Tag(name=name, **param_dict).create()
    return new_tag

